package ch.ethz.matsim.projects.astra_2018_002.mode_choice;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import org.matsim.api.core.v01.network.Network;
import org.matsim.core.config.groups.PlansCalcRouteConfigGroup;
import org.matsim.core.controler.AbstractModule;
import org.matsim.core.router.TripRouter;
import org.matsim.core.router.util.TravelTime;
import org.matsim.facilities.ActivityFacilities;
import org.matsim.pt.transitSchedule.api.TransitSchedule;
import org.opengis.geometry.MismatchedDimensionException;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.NoSuchAuthorityCodeException;
import org.opengis.referencing.operation.TransformException;

import com.fasterxml.jackson.core.JsonParseException;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.google.inject.Provides;
import com.google.inject.Singleton;
import com.google.inject.name.Named;

import ch.ethz.matsim.baseline_scenario.config.CommandLine;
import ch.ethz.matsim.baseline_scenario.config.CommandLine.ConfigurationException;
import ch.ethz.matsim.ch_pt_utils.costs.CostCalculator;
import ch.ethz.matsim.ch_pt_utils.costs.t603.SpatialT603;
import ch.ethz.matsim.ch_pt_utils.costs.t603.T603;
import ch.ethz.matsim.ch_pt_utils.costs.zonal.ZonalSystem;
import ch.ethz.matsim.mode_choice.constraints.CompositeTourConstraintFactory;
import ch.ethz.matsim.mode_choice.constraints.TourConstraintFromTripConstraint;
import ch.ethz.matsim.mode_choice.estimation.ModeAwareTripEstimator;
import ch.ethz.matsim.mode_choice.estimation.TourEstimatorFromTripEstimator;
import ch.ethz.matsim.mode_choice.estimation.TripEstimatorCache;
import ch.ethz.matsim.mode_choice.framework.ModeAvailability;
import ch.ethz.matsim.mode_choice.framework.ModeChoiceModel;
import ch.ethz.matsim.mode_choice.framework.ModeChoiceModel.FallbackBehaviour;
import ch.ethz.matsim.mode_choice.framework.tour_based.ActivityTourFinder;
import ch.ethz.matsim.mode_choice.framework.tour_based.TourBasedModel;
import ch.ethz.matsim.mode_choice.framework.tour_based.TourFinder;
import ch.ethz.matsim.mode_choice.framework.tour_based.constraints.TourConstraintFactory;
import ch.ethz.matsim.mode_choice.framework.tour_based.estimation.TourCandidate;
import ch.ethz.matsim.mode_choice.framework.tour_based.estimation.TourEstimator;
import ch.ethz.matsim.mode_choice.framework.trip_based.constraints.TripConstraintFactory;
import ch.ethz.matsim.mode_choice.framework.utilities.MultinomialSelector;
import ch.ethz.matsim.mode_choice.framework.utilities.UtilitySelectorFactory;
import ch.ethz.matsim.mode_choice.framework.utils.DefaultModeChainGenerator;
import ch.ethz.matsim.mode_choice.framework.utils.ModeChainGeneratorFactory;
import ch.ethz.matsim.mode_choice.mode_availability.LicenseAndOwnershipModeAvailability;
import ch.ethz.matsim.mode_choice.prediction.TeleportationPredictor;
import ch.ethz.matsim.mode_choice.replanning.ModeChoiceModelStrategy;
import ch.ethz.matsim.mode_choice.replanning.NonSelectedPlanSelector;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.constraints.AvoidOnlyWalkConstraint;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.constraints.VehicleTourConstraint;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.ModeChoiceParameters;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.car.CarEstimator;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.car.CarPredictor;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.other.BikeEstimator;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.other.WalkEstimator;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.pt.PublicTransportEstimator;
import ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.pt.PublicTransportPredictor;

public class ModeChoiceModule extends AbstractModule {
	private static final List<String> MODES = Arrays.asList("car", "pt", "bike", "walk");
	private static final List<String> VEHICLE_MODES = Arrays.asList("car", "bike");
	private static final List<String> CACHED_MODES = Arrays.asList("car", "pt", "bike", "walk");

	private final File ptPricingPath;
	private final CommandLine cmd;

	public ModeChoiceModule(File ptPricingPath, CommandLine cmd) {
		this.ptPricingPath = ptPricingPath;
		this.cmd = cmd;
	}

	@Override
	public void install() {
		addPlanStrategyBinding("custom").toProvider(ModeChoiceModelStrategy.class);
		bindPlanSelectorForRemoval().to(NonSelectedPlanSelector.class);
	}

	public final static String SCORING_PREFIX = "scoring";

	@Provides
	@Singleton
	public ModeChoiceParameters provideModeChoiceParameters() throws ConfigurationException, NoSuchFieldException,
			SecurityException, NumberFormatException, IllegalArgumentException, IllegalAccessException {
		List<String> scoringOptions = cmd.getAvailableOptions().stream().filter(o -> o.startsWith(SCORING_PREFIX + ":"))
				.collect(Collectors.toList());

		ModeChoiceParameters parameters = new ModeChoiceParameters();

		for (String option : scoringOptions) {
			String parameter = option.substring(SCORING_PREFIX.length() + 1);
			String value = cmd.getOptionStrict(option);

			Field field = ModeChoiceParameters.class.getField(parameter);
			field.set(parameters, Double.valueOf(value));
		}

		return parameters;
	}

	@Provides
	public ModeChoiceModel provideModeChoiceModel(PlansCalcRouteConfigGroup routeConfig,
			ModeChoiceParameters parameters, TripRouter router, ActivityFacilities facilities, Network network,
			@Named("car") TravelTime carTravelTime, CostCalculator costCalculator, TransitSchedule schedule)
			throws ConfigurationException {
		// SET UP ESTIMATORS

		// Walk estimator

		double crowflyDistanceFactorWalk = routeConfig.getModeRoutingParams().get("walk").getBeelineDistanceFactor();
		double speedWalk = routeConfig.getModeRoutingParams().get("walk").getTeleportedModeSpeed();

		TeleportationPredictor teleportationPredictorWalk = new TeleportationPredictor(crowflyDistanceFactorWalk,
				speedWalk);
		WalkEstimator walkEstimator = new WalkEstimator(parameters, teleportationPredictorWalk);

		// Bike estimator

		double crowflyDistanceFactorBike = routeConfig.getModeRoutingParams().get("bike").getBeelineDistanceFactor();
		double speedBike = routeConfig.getModeRoutingParams().get("bike").getTeleportedModeSpeed();

		TeleportationPredictor teleportationPredictorBike = new TeleportationPredictor(crowflyDistanceFactorBike,
				speedBike);
		BikeEstimator bikeEstimator = new BikeEstimator(parameters, teleportationPredictorBike);

		// Public transport estimator

		PublicTransportPredictor publicTransportPredictor = new PublicTransportPredictor(router, facilities, network,
				costCalculator, schedule);
		PublicTransportEstimator publicTransportEstimator = new PublicTransportEstimator(parameters,
				publicTransportPredictor);

		// Car estimator

		CarPredictor carPredictor = new CarPredictor(router, facilities, carTravelTime, network);
		CarEstimator carEstimator = new CarEstimator(parameters, carPredictor);

		// Create composite estimator

		ModeAwareTripEstimator modeAwareEstimator = new ModeAwareTripEstimator();
		modeAwareEstimator.addEstimator("walk", walkEstimator);
		modeAwareEstimator.addEstimator("bike", bikeEstimator);
		modeAwareEstimator.addEstimator("pt", publicTransportEstimator);
		modeAwareEstimator.addEstimator("car", carEstimator);

		TripEstimatorCache estimator = new TripEstimatorCache(modeAwareEstimator, CACHED_MODES);

		// Set up tour estimator
		TourEstimator tourEstimator = new TourEstimatorFromTripEstimator(estimator);

		// SET UP CONSTRAINTS

		// Trip constraints
		TripConstraintFactory avoidOnlyWalkConstraintFactory = new AvoidOnlyWalkConstraint.Factory();

		// Tour constraints
		TourConstraintFactory vehicleContinuityConstraintFactory = new VehicleTourConstraint.Factory(VEHICLE_MODES);

		// Combine constraints
		CompositeTourConstraintFactory constraintFactory = new CompositeTourConstraintFactory();
		constraintFactory.addFactory(new TourConstraintFromTripConstraint.Factory(avoidOnlyWalkConstraintFactory));
		constraintFactory.addFactory(vehicleContinuityConstraintFactory);

		// SET UP MODEL

		ModeAvailability modeAvailability = new LicenseAndOwnershipModeAvailability(MODES);
		ModeChainGeneratorFactory modeChainGeneratorFactory = new DefaultModeChainGenerator.Factory();
		TourFinder tourFinder = new ActivityTourFinder("home");

		UtilitySelectorFactory<TourCandidate> tourSelectorFactory = new MultinomialSelector.Factory<>(700.0);
		ModeChoiceModel model = new TourBasedModel(tourEstimator, modeAvailability, constraintFactory, tourFinder,
				tourSelectorFactory, modeChainGeneratorFactory, FallbackBehaviour.EXCEPTION);

		return model;
	}

	@Singleton
	@Provides
	public CostCalculator provideCostCalculator() throws JsonParseException, JsonMappingException, IOException,
			MismatchedDimensionException, NoSuchAuthorityCodeException, FactoryException, TransformException {
		File t603File = new File(ptPricingPath, "t603.json");
		File hafasPath = new File(ptPricingPath, "hafas");
		File zonesPath = new File(ptPricingPath, "zones/Tarifzonen NPVM LV95_area.SHP");

		T603 t603 = T603.read(t603File);
		SpatialT603 spatialT603 = SpatialT603.read(hafasPath, t603);
		ZonalSystem zonalPricing = ZonalSystem.read(zonesPath);

		return new CostCalculator(t603, spatialT603, zonalPricing);
	}
}
