package ch.ethz.matsim.projects.astra_2018_002.mode_choice.estimation.pt;

import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import org.matsim.api.core.v01.Coord;
import org.matsim.api.core.v01.network.Network;
import org.matsim.api.core.v01.population.Leg;
import org.matsim.api.core.v01.population.Person;
import org.matsim.core.router.LinkWrapperFacility;
import org.matsim.core.router.TripRouter;
import org.matsim.core.router.TripStructureUtils.Trip;
import org.matsim.core.utils.geometry.CoordUtils;
import org.matsim.facilities.ActivityFacilities;
import org.matsim.facilities.ActivityFacility;
import org.matsim.pt.transitSchedule.api.TransitLine;
import org.matsim.pt.transitSchedule.api.TransitRoute;
import org.matsim.pt.transitSchedule.api.TransitRouteStop;
import org.matsim.pt.transitSchedule.api.TransitSchedule;

import ch.ethz.matsim.baseline_scenario.transit.routing.EnrichedTransitRoute;
import ch.ethz.matsim.ch_pt_utils.costs.CostCalculator;
import ch.ethz.matsim.mode_choice.framework.ModeChoiceTrip;

public class PublicTransportPredictor {
	final private TripRouter router;
	final private ActivityFacilities facilities;
	final private Network network;

	final private CostCalculator costCalculator;
	final private TransitSchedule schedule;

	public PublicTransportPredictor(TripRouter router, ActivityFacilities facilities, Network network,
			CostCalculator costCalculator, TransitSchedule schedule) {
		this.router = router;
		this.facilities = facilities;
		this.network = network;
		this.costCalculator = costCalculator;
		this.schedule = schedule;
	}

	public PublicTransportPrediction predict(ModeChoiceTrip trip) {
		Trip tripInformation = trip.getTripInformation();

		ActivityFacility originFacility = facilities.getFacilities()
				.get(tripInformation.getOriginActivity().getFacilityId());
		ActivityFacility destinationFacility = facilities.getFacilities()
				.get(tripInformation.getDestinationActivity().getFacilityId());

		LinkWrapperFacility newOriginFacility = new LinkWrapperFacility(
				network.getLinks().get(tripInformation.getOriginActivity().getLinkId()));
		LinkWrapperFacility newDestinationFacility = new LinkWrapperFacility(
				network.getLinks().get(tripInformation.getDestinationActivity().getLinkId()));

		List<Leg> legs = router
				.calcRoute("pt", newOriginFacility, newDestinationFacility,
						tripInformation.getOriginActivity().getEndTime(), trip.getPerson())
				.stream().filter(Leg.class::isInstance).map(Leg.class::cast).collect(Collectors.toList());

		double crowflyDistance = CoordUtils.calcEuclideanDistance(originFacility.getCoord(),
				destinationFacility.getCoord());

		boolean isOnlyWalk = true;
		int numberOfTransfers = -1;

		double transferTime = 0.0;

		double inVehicleTime = 0.0;
		double inVehicleDistance = 0.0;

		double accessEgressTime = 0.0;

		for (Leg leg : legs) {
			if (leg.getMode().equals("pt")) {
				numberOfTransfers++;

				EnrichedTransitRoute route = (EnrichedTransitRoute) leg.getRoute();

				inVehicleTime += route.getInVehicleTime();
				inVehicleDistance += route.getDistance();

				if (isOnlyWalk) {
					// This is the first PT leg. Only score 60s of waiting time here!
					transferTime += Math.min(route.getWaitingTime(), 60.0);
				} else {
					transferTime += route.getWaitingTime();
				}

				isOnlyWalk = false;
			} else if (leg.getMode().contains("walk")) {
				accessEgressTime += leg.getTravelTime();
			} else {
				throw new IllegalStateException("Can only enrich pt and *_walk legs");
			}
		}

		numberOfTransfers = Math.max(0, numberOfTransfers);

		// double cost = calculateCost(trip.getPerson(), legs);
		double cost = calculateSimpleCost(trip.getPerson(), inVehicleDistance);
		
		return new PublicTransportPrediction(inVehicleTime, inVehicleDistance, accessEgressTime, transferTime,
				numberOfTransfers, isOnlyWalk, crowflyDistance, cost);
	}
	
	private double calculateSimpleCost(Person person, double inVehicleDistance) {
		boolean hasGA = (Boolean) person.getAttributes().getAttribute("ptHasGA");
		boolean hasHalbtax = (Boolean) person.getAttributes().getAttribute("ptHasHalbtax");

		if (hasGA) {
			return 0.0;
		} else {
			double cost = inVehicleDistance * 1e-3 * 0.25;
			return hasHalbtax ? 0.5 * cost : cost;
		}
	}

	private double calculateCost(Person person, List<Leg> legs) {
		boolean hasGA = (Boolean) person.getAttributes().getAttribute("ptHasGA");
		boolean hasHalbtax = (Boolean) person.getAttributes().getAttribute("ptHasHalbtax");

		if (hasGA) {
			return 0.0;
		} else {
			List<Coord> coords = new LinkedList<>();
			Coord lastCoord = null;

			for (Leg leg : legs) {
				if (leg.getMode().equals("pt")) {
					EnrichedTransitRoute route = (EnrichedTransitRoute) leg.getRoute();

					TransitLine transitLine = schedule.getTransitLines().get(route.getTransitLineId());
					TransitRoute transitRoute = transitLine.getRoutes().get(route.getTransitRouteId());

					TransitRouteStop originStop = transitRoute.getStops().get(route.getAccessStopIndex());
					TransitRouteStop destinationStop = transitRoute.getStops().get(route.getEgressStopIndex());

					coords.add(originStop.getStopFacility().getCoord());
					lastCoord = destinationStop.getStopFacility().getCoord();
				}
			}

			if (lastCoord != null) {
				coords.add(lastCoord);
			}

			return costCalculator.computeCost(coords, hasHalbtax);
		}
	}
}
